/************************************************************************* 
 
    Quick And Dirty Apricot Emulator
    Copyright (C) 2005, 2007  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "qdae.hxx"
#include "qdae_ui.hxx"
#include "sdl_video.hxx"
#include "fp_video.hxx"
#include "SdlContext.hxx"
#include "SdlTerm.hxx"
#include "sdl_keyboard.hxx"
#include "wd2797.hxx"

#define SCREEN_W 640
#define SCREEN_H 256
#define SCREEN_DEPTH 32
#define SCREEN_FLAGS 0
#define SCREEN_EXTRA 16
#define FPS_DELAY 100	/* Giving 10fps */

/* On a fully-populated FP, there are 4 pages of video RAM at 0xD0000,
 * 0xD8000, 0xE0000 and 0xE8000. Change bits 2 and 4 of port 2Eh to select
 * page displayed. 
 *
 * LCD displays from even addresses (0xE8000, 0xE8002, 0xE8004... )
 * Colour monitor presumably displays from odd addresses (0xE8001, ...)
 *
 * Bits 5 and 6 of port 2Eh control whether memory writes in the Exxxx
 * range affect (a) the low bytes, (b) the high bytes, (c) both.
 *
 * Bit 3 of port 2Eh can turn the LCD on or off.
 *
 * The colour monitor displays either 3 of the planes (for 8-colour) or all 4
 * (for 16-colour). Can also do 1-plane and 2-plane modes. 
 *
 * The palette is done by writing to port 2Ah; high nibble gives address
 * (bit 0: from D0000 plane. Bit 1: from D8000 plane...)
 *
 * Low nibble gives value. Bit 0 is blue, 1 is green, 2 is red, 3 intensity.
 *
 * Colour display uses 4 segments, like the Herc. There's an actual 6845 in
 * there, but its hardware cursor isn't used. 6Ch selects register, 6Eh value.
 *
 * Port 28h: LCD contrast, 0xFF = maximum
 *
 * Port 2Eh: Values written are 16-bit WORDS:
 * 		Bit 0 set: Reset CRTC
 * 		Bit 1:     LCD page base, bit 15
 * 		Bit 2:     Colour display enabled
 * 		Bit 3:     LCD enabled
 * 		Bit 4:     LCD page base, bit 16
 * 		Bit 5:     Set to enable access to low bytes in Exxxxx range
 * 		Bit 6:	   Set to enable access to high bytes in Exxxxx range
 * 		Bit 7:     Set to page video RAM in, clear to page it out
 * 		Bit 8:	   Not used
 * 		Bit 9:	   ~STOP LED
 * 		Bit 10:	   ~POWER LED
 * 		Bit 11:	   ~SHIFT LOCK LED
 * 		Bit 12:	   ~DISK LED
 * 		Bit 13:    ~VOICE LED
 * 		Bit 14:	   ~COLOUR SELECT LED
 * 		Bit 15:	   ~CAPS LOCK LED
 */
#define VRAMBASE 0xE8000
#define VRAMLEN  0x1000
#define PALETTE 0xE0000
#define VPTRBASE 0x1E00
#define VPTRLEN  0x1000

//#define VIDEO_LOG(x)  printf x ; fflush(stdout);

#define VIDEO_LOG(x)


FP_Video::FP_Video(PC_ENV *m) : Video(m, "fp", SCREEN_W, SCREEN_H * 2 + SCREEN_EXTRA)
{
	extra_height = SCREEN_EXTRA;
	vscale = 2;
	scr_rows = 256;
}

FP_Video::~FP_Video()
{
}

u_int8 FP_Video::read(u_int16 port)
{
	switch(port)
	{
		case 0x6C: return m_crtc.m_cur;
		case 0x6E: return m_crtc.get();
	}
	return 0xff;
}

void FP_Video::write16(u_int16 port, u_int16 value)
{
	if (port == 0x2E)
	{

	}
	else write(port, value);
}

void FP_Video::write(u_int16 port, u_int8 value)
{
	switch(port)
	{
		case 0x6C: m_crtc.m_cur = value; break;
		case 0x6E: 
   			VIDEO_LOG(("Video: CRTC[%d] = %d\n", 
					m_crtc.m_cur, value));
			m_crtc.set(value);
/* XXX Possibly take account of some values written */
			break;

	}
#if 0
		case 13: /* Set Stop LED */
			gl_stop = value;
			gl_ledsdirty = 1;
			break;

		case 11: /* Set Caps Lock LED */
			gl_caps = value;
			gl_ledsdirty = 1;
			break;

		case 9: /* Set horizontal resolution */
			v_mode = value;
			gl_dirty = 1;
			break;

		case 7: /* Set vertical resolution */
			gl_dirty = 1;
			if (value == 0) 
			{
				SDL_Rect rc;
	
				scr_rows = 200;
				rc.x = 0;
				rc.y = 200 * vscale;
				rc.w = SCREEN_W;
				rc.h = (SCREEN_H - 200) * vscale;
				SDL_FillRect(vidSurface, &rc, colours[15]);
			}
			else	scr_rows = 256;
			break;
	}
#endif
}

#if 0
void FP_Video::draw_row_2bit(int y, unsigned char *row, unsigned char *shadow,
			unsigned char *palette)
{
	int y1, x1, x;
	unsigned char mask;
	unsigned char *font;
	unsigned char *framebuf;
	unsigned offset;
	unsigned char *bitmap;
	unsigned bpp;
	int col;

	bpp = vidSurface->format->BytesPerPixel;
       	framebuf = (unsigned char *)(vidSurface->pixels);
	for (y1 = 0; y1 < vscale; y1++)
	{
		offset = ((y * vscale + y1) * vidSurface->pitch);

		for (x = 0; x < 160; x += 2, offset += 8 * bpp)
		{
			if (shadow && (row[x] == shadow[x]) &&
					(row[x+1] == shadow[x+1])) continue;
			mask = 0x80;
			for (x1 = 0; x1 < 8; x1++)
			{
				u_int8 d0 = row[x]   & mask;
				u_int8 d1 = row[x+1] & mask;
				if (d0 && d1) col = 10;
				else if (d1)  col = 8;
				else if (d0)  col = 2;
				else          col = 0;
				*(Uint32 *)(&framebuf[offset + x1 * bpp]) = 
					colours[palette[col*2] & 0xF];
				mask = mask >> 1;
			}
		}
	}
}	



void FP_Video::draw_row_4bit(int y, unsigned char *row, unsigned char *shadow,
			unsigned char *palette)
{
	int y1, x1, x;
	unsigned char mask;
	unsigned char *font;
	unsigned char *framebuf;
	unsigned offset;
	unsigned char *bitmap;
	unsigned bpp;
	int col;

	bpp = vidSurface->format->BytesPerPixel;
       	framebuf = (unsigned char *)(vidSurface->pixels);
	for (y1 = 0; y1 < vscale; y1++)
	{
		offset = ((y * vscale + y1) * vidSurface->pitch);

		for (x = 0; x < 160; x += 2, offset += 8 * bpp)
		{
			if (shadow && (row[x] == shadow[x]) &&
					(row[x+1] == shadow[x+1])) continue;
			mask = 0xC0;
			for (x1 = 0; x1 < 4; x1++)
			{
				u_int8 d0 = row[x]   & mask;
				u_int8 d1 = row[x+1] & mask;

				switch(x1)
				{
					case 0: col = (d0 >> 6) | (d1 >> 4); 
						break;
					case 1: col = (d0 >> 4) | (d1 >> 2); 
						break;
					case 2: col = (d0 >> 2) | (d1); 
						break;
					case 3: col = (d1 << 2) | (d0);
						break;
				}
				*(Uint32 *)(&framebuf[offset + (2*x1) * bpp]) = 
					colours[palette[col*2] & 0xF];
				*(Uint32 *)(&framebuf[offset + (2*x1+1)*bpp]) = 
					colours[palette[col*2] & 0xF];
				mask = mask >> 2;
			}
		}
	}
}	
#endif

void FP_Video::draw_row_lcd(int y, unsigned char *row, unsigned char *shadow)
{
	int y1, x1, x;
	unsigned char mask;
	unsigned char *font;
	unsigned char *framebuf;
	unsigned offset;
	unsigned char *bitmap;
	unsigned bpp;
	int col;

	bpp = vidSurface->format->BytesPerPixel;
       	framebuf = (unsigned char *)(vidSurface->pixels);
	for (y1 = 0; y1 < vscale; y1++)
	{
		offset = ((y * vscale + y1) * vidSurface->pitch);

		for (x = 0; x < 160; x += 2, offset += 8 * bpp)
		{
			if (shadow && (row[x] == shadow[x])) continue;
			mask = 0x80;
			for (x1 = 0; x1 < 8; x1++)
			{
				u_int8 d0 = row[x]   & mask;
				*(Uint32 *)(&framebuf[offset + x1 * bpp]) = 
					d0 ? m_lcdfg : m_lcdbg;
				mask = mask >> 1;
			}
		}
	}
}


int FP_Video::paint()
{
	int y;
	int any;
	long offset, shadow_offset;
	SDL_Event e;
	gl_dirty = 1;
	static unsigned char shadow_palette[32];

	for (;;)
	{
		SDL_Delay(FPS_DELAY);

		// If emulator UI is using the screen, don't paint.
		if (gl_ui->usingScreen()) continue;

#if 0 // We don't support the palette yet
		if (memcmp(sysram + PALETTE, shadow_palette, 32))
		{
			gl_dirty = 1;
			memcpy(shadow_palette, sysram + PALETTE, 32);	
/*			printf("New palette: ");
			for (y = 0; y < 32; y+=2)
			{
				if (shadow_palette[y] != sysram[y + PALETTE])
				{
					printf("<%02x>", sysram[y + PALETTE]);

				}
				else printf(" %02x ", sysram[y + PALETTE]);
			}
			printf("\n"); */
		}
#endif
		any = 0;

/* XXX There are two cases to consider:
 * i) We're drawing the LCD.
 * ii) We're drawing the external monitor
 *
 * For now, just do the LCD because it's easy */

		for (y = 0; y < 200 /* scr_rows */ /* 256 */; y++)
		{
			/* XXX Force page 3 */
			offset = 0x18000 + 160 * y;
			if (gl_dirty)
			{
				if (!any) SDL_LockSurface(vidSurface);
				any = 1;
				draw_row_lcd(y, videoram + offset, NULL);
			}
			else if (memcmp(videoram + offset, shadowram + offset, 
						160))
			{
				if (!any) SDL_LockSurface(vidSurface);
				any = 1;
				draw_row_lcd(y, videoram + offset, 
						/*NULL */ shadowram + offset);
				memcpy(shadowram + offset, videoram + offset, 
						160);
			}
			/* else  // Force drawing
			{
				if (!any) SDL_LockSurface(vidSurface);
				any = 1;
				draw_row(y, sysram + offset, NULL, sysram + PALETTE);
			} */
		}
		if (any)
		{
			SDL_UpdateRect(vidSurface, 0, 0, 0, 0);
			SDL_UnlockSurface(vidSurface);
		}
		if (gl_dirty || gl_ledsdirty)
		{
			term->setForeground(0, 0xFF, 0);
			term->moveCursor(0, 0);
			(*term) << (gl_caps ? "CAPS " : "     ");
			term->setForeground(0xFF, 0xFF, 0);
			(*term) << (gl_stop ? "STOP " : "     ");
			term->setForeground(0xFF, 0x00, 0);
			(*term) << (floppy_indicator[0] ? "A " : "  ");
			(*term) << (floppy_indicator[1] ? "B " : "  ");
			gl_ledsdirty = 0;
		}
		gl_dirty = 0;
	}
}

/* Fairly arbitrary 16-colour palette, in the same order as the
 * Rascal C (F-series)  */
static int r[16] = {  255, 191,   0,   0, 255, 191,   0,   0,
		      255, 191,   0,   0, 255, 191,  63,   0};

static int g[16] = {  255, 191, 255, 191,   0,   0,   0,   0,
		      255, 191, 255, 191,   0,   0,  63,   0 };

static int b[16] = {  255, 191, 255, 191, 255, 191, 255, 191,
     		        0,   0,   0,   0,   0,   0,  63,   0};


void FP_Video::init()
{
	int n;

	Video::init();

	memset(videoram, 0, sizeof(videoram));
	memset(shadowram, 0, sizeof(shadowram));
	m_lcdbg = SDL_MapRGB(vidSurface->format, 0x71, 0xa7, 0xbf);
	m_lcdfg = SDL_MapRGB(vidSurface->format, 0x3a, 0x5d, 0x6e);

	for (n = 0; n < 16; n++)
	{
		m_palette[n] = n;
		colours[n] = SDL_MapRGB(vidSurface->format, r[n], g[n], b[n]);
	}
}

/* XXX We may need to support the "base model FP" case where the
 * only VRAM is 16k at even addresses between E8000 and EFFFE, by
 * rigging fetch() or store(). Not to mention the whole
 * 'enable odd/even addresses' thing. */

void FP_Video::store(u_int32 offset, u_int8 value)
{
	offset &= (VRAMSIZE - 1);
	if (offset == 0)
	{
		fprintf(stderr, "Video: POKE 0, 0x%02x\n", value);
	}
	videoram[offset] = value;
}


u_int8 FP_Video::fetch(u_int32 offset)
{
	offset &= (VRAMSIZE - 1);
	return videoram[offset];
}
