/************************************************************************

    QDAE v0.0.2 - Quick And Dirty Apricot emulator

    Copyright (C) 2005-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "qdae.hxx"
#include "qdae_ui.hxx"
#include "qdae_files.hxx"
#include "SdlContext.hxx"
#include "SdlTerm.hxx"
#include "UiContainer.hxx"
#include "UiLabel.hxx"
#include "UiCommand.hxx"
#include "UiSetting.hxx"
#include "UiMenu.hxx"
#include "UiMenuBar.hxx"
#include "UiMenuBarCommand.hxx"
#include "UiScrollingMenu.hxx"
#include "UiSeparator.hxx"
#include "UiFileEntry.hxx"
#include "UiFileList.hxx"
#include "UiFileChooser.hxx"
#include "Path.hxx"
#include "wd2797.hxx"
#include "sdl_video.hxx"

#define MENU_X(ww)  (*m_menuBar)[m_menuBar->getSelected()].getX() \
                 + ((*m_menuBar)[m_menuBar->getSelected()].getWidth() / 2) \
                            - (ww * charW / 2)

#define UI_CHAR_W 8
#define UI_CHAR_H 16


static const char *dsktypes[] =
{
        NULL,           // 0
        "floppy",       // 1
        "dsk",          // 2
        "raw",          // 3
        "myz80",        // 4
        "cfi",          // 5
        "edsk",         // 6
        "ntwdm",        // 7
        "rcpmfs",       // 8
        "apridisk",     // 9
};


static inline int  alignX(int x)
{
	return x - (x % UI_CHAR_W);
}


static inline int  alignY(int y)
{
	return y - (y % UI_CHAR_H);
}




QDAE_UI::QDAE_UI(PC_ENV *host, SDL_Surface *s)
{
	m_host = host;
	m_term = new SdlTerm(s);
	m_surface = s;
	m_nUsingScreen = 0;
	m_clrFG.r = 0; m_clrFG.g = 0; m_clrFG.b = 0;
	m_clrBG.r = 0xE0; m_clrBG.g = 0xFF; m_clrBG.b = 0xFF;
	m_menuBar = new UiMenuBar(this);
	m_menuBar->add(new UiMenuBarCommand(SDLK_ESCAPE, "ESC=Back to QDAE", this));
	m_menuBar->add(new UiMenuBarCommand(SDLK_F1, "F1=Help", this));
	m_menuBar->add(new UiMenuBarCommand(SDLK_F2, "F2=Drives", this));
	m_menuBar->add(new UiMenuBarCommand(SDLK_F3, "F3=Mouse", this));
	m_menuBar->add(new UiMenuBarCommand(SDLK_F5, "F5=Reset", this));
	m_menuBar->add(new UiMenuBarCommand(SDLK_F6, "F6=Settings", this));
	m_menuBar->add(new UiMenuBarCommand(SDLK_F7, "F7=Dbg", this));
	m_menuBar->add(new UiMenuBarCommand(SDLK_F10, "F10=Quit", this));
}

QDAE_UI::~QDAE_UI()
{
	delete m_menuBar;
	delete m_term;
}

void QDAE_UI::acquireScreen()
{
	++m_nUsingScreen;
}


void QDAE_UI::releaseScreen()
{
	if (m_nUsingScreen > 0) --m_nUsingScreen;
}


/* UiDrawer overrides */
void QDAE_UI::measureString(const string &s, Uint16 *w, Uint16 *h)
{
	*w = UI_CHAR_W * s.size();
	*h = UI_CHAR_H;
}


void QDAE_UI::measureMenuBorder(SDL_Rect &rcInner, SDL_Rect &rcOuter)
{
	rcOuter = rcInner;
	rcOuter.x -= UI_CHAR_W;
	rcOuter.y -= UI_CHAR_W;
	rcOuter.w += 2 * UI_CHAR_W;
	rcOuter.h += 2 * UI_CHAR_H;
}

void QDAE_UI::measureMenuBar(SDL_Rect &rc)
{
	rc.h = UI_CHAR_H;
	rc.w = m_surface->w;
	rc.y = UI_CHAR_H;
	rc.x = 0;
}

void QDAE_UI::drawRect   (int selected, SDL_Rect &rc, SDL_Colour *c)
{
	if (c)
	{
		Uint32 oldfg = m_term->getForeground();
		m_term->setForeground(*c);
		m_term->fillRect(rc, m_term->getForeground());
		m_term->setForeground(oldfg);
	}
	else if (selected) m_term->fillRect(rc, m_term->getForeground());
	else	       m_term->fillRect(rc, m_term->getBackground());
}


void QDAE_UI::drawString (int selected, Uint16 x, Uint16 y, const std::string s)
{
	x /= UI_CHAR_W;
	y /= UI_CHAR_H;

	if (selected) m_term->setReverse(true);
	else	  m_term->setReverse(false);

	m_term->moveCursor(x, y);
	(*m_term) << s.c_str();
}


void QDAE_UI::drawPicture(int selected, Uint16 x, Uint16 y, SDL_Surface *s)
{
	SDL_Rect rcFrom, rcTo;

	rcFrom.x = rcFrom.y = 0;
	rcFrom.w = s->w;
	rcFrom.h = s->h;
	rcTo.x = x;
	rcTo.y = y;
	rcTo.w = s->w;
	rcTo.h = s->h;

	SDL_BlitSurface(s, &rcFrom, m_surface, &rcTo);
}

void QDAE_UI::drawGlyph  (int selected, Uint16 x, Uint16 y, UiGlyph g)
{
	unsigned char ch;

	x /= UI_CHAR_W;
	y /= UI_CHAR_H;

	if (selected) m_term->setReverse(true);
	else	  m_term->setReverse(false);

	m_term->moveCursor(x, y);
	switch(g)
	{
		case UIG_NONE:	  ch = ' '; break;
		case UIG_CHECKED:       ch = 188; break;
		case UIG_UNCHECKED:     ch = 187; break;
		case UIG_COMMAND:       ch = 252; break;
		case UIG_SUBMENU:       ch = 128; break;
	}
	(*m_term) << ch;
}

void QDAE_UI::drawMenuBorder(SDL_Rect &rcOuter, SDL_Rect &rcInner)
{
	SDL_Rect rc;

	m_term->setForeground(m_clrFG);
	m_term->setBackground(m_clrBG);

	memcpy(&rc,       &rcOuter, sizeof(SDL_Rect));
	memcpy(&rcInner,  &rcOuter, sizeof(SDL_Rect));

	m_term->fillRect(rc, m_term->getBackground());
	m_term->rect    (rc, m_term->getForeground());

	rcInner.x += UI_CHAR_W;
	rcInner.y += UI_CHAR_H;
	rcInner.w -= 2 * UI_CHAR_W;
	rcInner.h -= 2 * UI_CHAR_H;
}

void QDAE_UI::drawMenuBar   (int selected, SDL_Rect &rc)
{
	m_term->setForeground(m_clrFG);
	m_term->setBackground(m_clrBG);
	drawRect(selected, rc);
}


void QDAE_UI::drawSeparator (int selected, SDL_Rect &rc)
{
	rc.y += 6;
	rc.h  = 2;
	rc.w += 2 * UI_CHAR_W;
	rc.x -= UI_CHAR_W;
	drawRect(!selected, rc);
}


void QDAE_UI::drawCursor    (Uint16 x, Uint16 y, char ch)
{
	x /= UI_CHAR_W;
	y /= UI_CHAR_H;
		 
	m_term->setReverse(false);
	m_term->moveCursor(x, y);
	(*m_term) << (unsigned char)ch;
}

bool QDAE_UI::getManualUpdate(void)
{
	return m_term->getManualUpdate();
}


void QDAE_UI::setManualUpdate(bool b)
{
	m_term->setManualUpdate(b);
}


void QDAE_UI::updateRect(SDL_Rect &rc)
{
	SDL_UpdateRects(m_surface, 1, &rc);
}

SDL_Surface *QDAE_UI::saveSurface(SDL_Rect &rc)
{
	SDL_Rect rcto;
	SDL_Surface *s2;

	rcto.x = rcto.y = 0;
	rcto.w = rc.w;
	rcto.h = rc.h;

	s2 = SDL_AllocSurface(SDL_SWSURFACE, rcto.w, rcto.h,
		m_surface->format->BitsPerPixel, 0, 0, 0, 0);

	if (!s2) return NULL;
// XXX	SDL_SetColours(s2, m_sysVideo->m_colours, 0, 256);
	SDL_BlitSurface(m_surface, &rc, s2, &rcto);
	return s2;
}

void QDAE_UI::restoreSurface(SDL_Surface *s, SDL_Rect &rc)
{
	SDL_BlitSurface(s, NULL, m_surface, &rc);
	SDL_FreeSurface(s);
	SDL_UpdateRects(m_surface, 1, &rc);
}


void QDAE_UI::centreContainer(UiContainer &c)
{
	unsigned int sx, sy;

	c.pack();

	sx = ((m_surface->w - c.getWidth ()) / 2);
	sy = ((m_surface->h - c.getHeight()) / 2);

	c.setX(alignX(sx));
	c.setY(alignY(sy));
}

UiEvent QDAE_UI::menuEvent(SDL_Event &e)
{
	int r = m_term->onEvent(e);

	switch(r)
	{
	      case -2: return UIE_QUIT;
	      case 0:  return UIE_CONTINUE;
	      case 1:  return UIE_OK;
	}
	return UIE_CONTINUE;
}


char QDAE_UI::getKey(void)
{
	if (!m_term->kbhit()) return 0;
	return m_term->read();
}


void QDAE_UI::onTimer50(void)
{
	m_term->onTimer50();
}

UiEvent QDAE_UI::reportV(const string s, const char *prompt, ...)
{
	char buf[78];	// Screen width, more or less
	va_list ap;

	va_start(ap, prompt);
#ifdef HAVE_VSNPRINTF
	vsnprintf(buf, sizeof(buf), s.c_str(), ap);
#else
	vsprintf(buf, s.c_str(), ap);
#endif
	va_end(ap);
	return report(buf, prompt);
}

UiEvent QDAE_UI::report(const string s, const string prompt)
{
	if (this == NULL)
	{
		fprintf(stderr, "Error starting QDAE: %s\n", s.c_str());
		exit(1);
	}
	UiMenu uim(this);
	SDL_Rect rc;
	int x,y;
	Uint16 sw, sh, bw, bh;

	rc.x = rc.y = 0;
	rc.w = m_surface->w;
	rc.h = m_surface->h;

	measureString(s,      &sw, &sh);
	measureString(prompt, &bw, &bh);
	if (bw > sw) sw = bw;
	if (bh > sh) sh = bh;
	measureString("A", &bw, &bh);

	uim.add(new UiLabel(s, this))
	   .add(new UiSeparator(this))
	   .add(new UiCommand(SDLK_o, prompt, this));
	y = (rc.h - 6 * bh) / 2;
	x = (rc.w - bw - sw - bw) / 2;
	x = alignX(x);
	y = alignY(y);
	uim.setX(x);
	uim.setY(y);
	acquireScreen();
	UiEvent e = uim.eventLoop();
	releaseScreen();
	return e;
}


UiEvent QDAE_UI::mainMenu()
{
	bool inMenu;
	int y;
	Uint16 charW, charH;
	UiEvent ev = UIE_CONTINUE;

	acquireScreen();
	menuDraw();
	SDL_ShowCursor(1);
        measureString("A", &charW, &charH);
	
	do
	{
		inMenu = true;
		ev = m_menuBar->eventLoop();
		y = m_menuBar->getY() + m_menuBar->getHeight();
		if (ev == UIE_QUIT) return ev;
		switch(m_menuBar->getKey(m_menuBar->getSelected()))
		{
/*
			case SDLK_F8:
				ev = actionMenu(MENU_X(21), y);
				if (ev == UIE_QUIT ||ev == UIE_OK) 
					inMenu = false;
				break; */
			case SDLK_ESCAPE:
				inMenu = false;
				break;

			case SDLK_F1:
				ev = menuF1(MENU_X(19), y);
				if (ev == UIE_QUIT)
					inMenu = false;
				break;
			case SDLK_F2:
				ev = menuF2(MENU_X(19), y);
				if (ev == UIE_QUIT)
					inMenu = false;
				break;
			case SDLK_F3:
				ev = menuF3(MENU_X(19), y);
				if (ev == UIE_QUIT)
					inMenu = false;
				break;
			case SDLK_F5:
				ev = confirm(MENU_X(17), y, "");
				if (ev == UIE_QUIT)
					inMenu = false;
				if (ev == UIE_OK)
				{
					inMenu = false;
					m_host->reboot();
				}
				break;
			case SDLK_F6:
				ev = menuF6(MENU_X(19), y);
				if (ev == UIE_QUIT)
					inMenu = false;
				break;
			case SDLK_F7:
				ev = menuF7(MENU_X(19), y);
				if (ev == UIE_QUIT)
					inMenu = false;
				break;
			case SDLK_F10:
				ev = confirm(m_surface->w - 17 * charW, y, "");
				if (ev == UIE_QUIT ||ev == UIE_OK) 
				{
					inMenu = false;
					ev = UIE_QUIT;
				}
				menuDraw();
				break;

		}		
	} while (inMenu);

	SDL_ShowCursor(!m_host->m_video->isMouseCaptured());
	releaseScreen();
	gl_dirty = 1;
	return ev;
}


void QDAE_UI::menuDraw()
{
	char menustr[101];

	m_term->setForeground(m_clrFG);
	m_term->setBackground(m_clrBG);
	m_term->fillRect(0, 0, m_surface->w, 2 * UI_CHAR_H, 
			m_term->getBackground());
	sprintf(menustr, "QDAE v" QDAE_VERSION "  Copyright 2005-6,2008 John Elliott"
			"  Compiled with SDL %d.%d.%d",
			SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_PATCHLEVEL);
	m_term->setReverse(false);
	m_term->home();
	(*m_term) << menustr;
}


UiEvent QDAE_UI::confirm(int x, int y, const string prompt)
{
	int lpad, rpad, tlen;
	string title;
	UiEvent e;
	UiMenu mc(this);

	x = alignX(x);
	if (prompt != "")
	{
		tlen = prompt.size();
		if (tlen >= 14) title = prompt.substr(0, 15);
		else
		{
			title = "";
			lpad = (15 - tlen) / 2;
			rpad = 15 - (tlen + lpad);
			while (lpad--) title += " ";
			title += prompt;
			while (rpad--) title += " ";
		}
		mc.add(new UiLabel(title, this));
	}
	mc.add(new UiLabel( "Are you sure? ", this))
	  .add(new UiSeparator(this))
	  .add(new UiCommand(SDLK_n, "  No", this))
	  .add(new UiCommand(SDLK_y, "  Yes", this));
	mc.setX(x);
	mc.setY(y);

	e = mc.eventLoop();
	if (e == UIE_QUIT) return e;
	if (mc.getKey(mc.getSelected()) == SDLK_y) return UIE_OK;
	return UIE_CANCEL;
}


UiEvent QDAE_UI::menuF1(int x, int y)
{
	UiMenu uim(this);
	UiEvent uie;

        x = alignX(x);
        y = alignY(y);

	uim.add(new UiCommand (SDLK_a, "  About QDAE...   ", this, UIG_SUBMENU));
//	uim.add(new UiCommand (SDLK_k, "  Keyboard layout  ", this));
	uim.setX(x);
	uim.setY(y);

	uie = uim.eventLoop();

	if (uie == UIE_QUIT) return uie;
	switch(uim.getKey(uim.getSelected()))
	{
		case SDLK_a: return aboutBox();
//		case SDLK_k: 
//			drawKeyboard();
		default:
			break;
	}
	return UIE_CONTINUE;
}


UiEvent QDAE_UI::aboutBox(void)
{
	UiMenu uim(this);

        uim.add(new UiLabel("This program is free software; you can redistribute it and/or modify it under", this))
	   .add(new UiLabel("the terms of the GNU General Public License as published by the Free Software", this))
	   .add(new UiLabel("Foundation; either version 2 of the License, or (at your option) any later", this))
	   .add(new UiLabel("version.", this))
	   .add(new UiLabel("", this))
	   .add(new UiLabel("This program is distributed in the hope that it will be useful, but WITHOUT", this))
	   .add(new UiLabel("ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or", this))
	   .add(new UiLabel("FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License ", this))
	   .add(new UiLabel("for more details.", this))
	   .add(new UiLabel("", this))
	   .add(new UiLabel("You should have received a copy of the GNU General Public License along with ", this))
	   .add(new UiLabel("this program; if not, write to the Free Software Foundation, Inc., 59 Temple", this))
	   .add(new UiLabel("Place, Suite 330, Boston, MA  02111-1307  USA" , this))
	   .add(new UiLabel("", this))
           .add(new UiCommand(SDLK_o, "  OK  ", this));

	centreContainer(uim);
	return uim.eventLoop();
}

/*
#include "pcwkey.xbm"	// X-Window bitmap, a graphic format designed to
			// be used in C source.

void JoyceMenuTermPcw::drawKeyboard(void)	// Draw the keyboard map 
{
	int y,x,b,z=0;
	byte ch;
	Uint32 fg, bg;

	if (!m_term->lock()) return;
	fg = m_term->getForeground();
        bg = m_term->getBackground();

	for (y = 0; y < pcwkey_height; y++)
		for (x = 0; x < pcwkey_width; x+=8)
		{
			ch = pcwkey_bits[z++];
			for (b = 0; b < 8; b++)
			{
				m_term->dot(x+b, y+32, (ch & 1) ? fg : bg);
				ch = (ch >> 1);
			}
		}
	m_term->unlock();
	SDL_UpdateRect(m_term->getSurface(),
			0, 32, pcwkey_width, pcwkey_height);

}
*/
UiEvent QDAE_UI::menuF2(int x, int y)
{
	int n;
	int sel = -1;
	UiEvent uie;
	static const char *driveName[] = {
		"  Drive A: (",
		"  Drive B: (",
		"  Winchester 0 (",
		"  Winchester 1 (" };

        x = alignX(x);
        y = alignY(y);
	while (1)
	{
		UiMenu uim(this);

		for (n = 0; n < MAX_FLOPPIES; n++)
		{
			string title = driveName[n];
			if (*floppy_filename[n] == "")
			{
				title += "No disc)  ";
				uim.add(new UiLabel(title, this));
				uim.add(new UiCommand((SDLKey)(SDLK_i + n), 
						"  Insert...  ", this));
			}
			else
			{
				title += displayName(*floppy_filename[n], 40);
				title += ")  ";
				uim.add(new UiLabel(title, this));
				uim.add(new UiCommand((SDLKey)(SDLK_e + n), 
						"  Eject  ", this));
			}
			uim.add(new UiSetting((SDLKey)(SDLK_r + n),
				floppy_readonly[n], "  Read only  ", this));
			uim.add(new UiSeparator(this));
		}
			//uim.add(new UiCommand(SDLK_m, "  Disc management...  ", d, UIG_SUBMENU));
			//uim.add(new UiSeparator(this));
		uim.add(new UiCommand(SDLK_ESCAPE, "  EXIT  ", this));
		uim.setX(x);
		uim.setY(y);
		uim.setSelected(sel);
	
		uie = uim.eventLoop();

		if (uie == UIE_QUIT) return uie;

// We have enough keycodes here for four drives - either four floppies, or 
// two floppies and two Winchesters. Except it doesn't make sense to 
// eject a Winchester!
		sel = uim.getSelected();
		switch(uim.getKey(sel))
		{
			case SDLK_e: m_host->floppy_eject(0); break;
			case SDLK_f: m_host->floppy_eject(1); break;
//			case SDLK_g: m_host->floppy_eject(2); break;
//			case SDLK_h: m_host->floppy_eject(3); break;

			case SDLK_i:
			case SDLK_j:
//			case SDLK_k:
//			case SDLK_l:
			{
				string curName;
				const char *type = floppy_type[n];

				n = uim.getKey(sel) - SDLK_i;
				curName = *floppy_filename[n];
				if (curName == "") 
					curName = findQdaeDir(FT_DISC, false);
				uie = requestDisc(curName, &type);
				if (uie == UIE_OK)
				{
					uie = m_host->floppy_insert(n, curName, type);
				}
			}
			break;
			case SDLK_r: floppy_readonly[0] = !floppy_readonly[0];
				     break;
			case SDLK_s: floppy_readonly[1] = !floppy_readonly[1];
				     break;
//			case SDLK_t: floppy_readonly[2] = !floppy_readonly[2];
//			case SDLK_t: winch_readonly[0]  = !winch_readonly[0];
//				     break;
//			case SDLK_u: floppy_readonly[3] = !floppy_readonly[3];
//			case SDLK_u: winch_readonly[1]  = !winch_readonly[1];
//				     break;
			case SDLK_ESCAPE:
				     return UIE_OK;
			default: break;
		}
	}
	return UIE_CONTINUE;
}


UiEvent QDAE_UI::requestDisc(string &filename, const char **type)
{
	int sel = -1;
	bool a,b,dsk,folder;
	int fmt = 0;
	UiEvent uie;
	int inMenu = 1;
	
	while(inMenu)
	{
		UiMenu uim(this);

		dsk = false;
		folder = false;
		if (filename == A_DRIVE) a = true; else a = false;
		if (filename == B_DRIVE) b = true; else b = false;
		if (!(a || b)) 
		{
			if (Path::isDir(filename)) folder = true;
			else	dsk = true; 
		}

		uim.add(new UiLabel("  Enter filename for disc image  ", this));
		uim.add(new UiSeparator(this));
		uim.add(new UiLabel  ("  File/drive: " + displayName(filename, 40) + "  ", this));
		uim.add(new UiSetting(SDLK_a, a, "  Drive A: [" A_DRIVE "]  ",this));
		uim.add(new UiSetting(SDLK_b, b, "  Drive B: [" B_DRIVE "]  ",this));
		uim.add(new UiSetting(SDLK_d, dsk,    "  Disc file...  ",this));
// JOYCE lets you mount folders as drives. Can't do this in QDAE until LibDsk
// supports doing it in MS-DOS format.
//		uim.add(new UiSetting(SDLK_f, folder, "  Folder...  ",this));
		uim.add(new UiSeparator(this));
		uim.add(new UiCommand(SDLK_v, "  Advanced...  ", this, UIG_SUBMENU));
		uim.add(new UiSeparator(this));
		uim.add(new UiCommand(SDLK_o, "  OK  ", this));
		uim.add(new UiCommand(SDLK_c, "  Cancel  ", this));
		uim.setSelected(sel);
                centreContainer(uim);
		uie = uim.eventLoop();
		if (uie == UIE_QUIT || uie == UIE_CANCEL) return uie;
		sel = uim.getSelected();
		switch(uim.getKey(sel))
		{
			case SDLK_a: filename = A_DRIVE; fmt = 1; break;
			case SDLK_b: filename = B_DRIVE; fmt = 1; break;
			case SDLK_c: return UIE_CANCEL;  break;
			case SDLK_o: inMenu = 0; break;
			case SDLK_d: 
                                {
                                    UiFileChooser f("  OK  ", this);
                                    if (dsk)
                                            f.initDir(findQdaeDir(FT_DISC, false));
                                    else    f.initDir(filename);

                                    uie = f.eventLoop();
                                    if (uie == UIE_QUIT) return uie;
                                    if (uie == UIE_OK) 
				    {
					filename = f.getChoice();
					fmt = 9;	
				    }
                                }
                                break;
/*
			case SDLK_f: 
                                {
                                    PcwFolderChooser f("  OK  ", this);
                                    if (!folder)
                                            f.initDir(findPcwDir(FT_DISC, false));
                                    else    f.initDir(filename);

                                    uie = f.eventLoop();
                                    if (uie == UIE_QUIT) return uie;
                                    if (uie == UIE_OK) 
				    {
					filename = f.getChoice();
					string ini = filename + "/.libdsk.ini";
					FILE *fp = fopen(ini.c_str(), "r");
					if (!fp)
					{
						int fmt;
						uie = getFolderType(d, fmt);
						if (uie == UIE_QUIT) return uie;
						if (uie == UIE_OK)
						{
							formatFolder(filename, fmt);
						}
					}
					else fclose(fp);
					fmt = 8;	// Folder
				    }
                                } */
                                break;
			case SDLK_v: 
				uie = getDiscType(fmt, false);
                                if (uie == UIE_QUIT) return uie;
				break;
			default: break;
		}
	}
	// A filename has been entered. See if we can open the matching
	// DSK.
	DSK_PDRIVER drv;
	dsk_err_t err = dsk_open(&drv, filename.c_str(), dsktypes[fmt], NULL);
	// If we failed to open a file as Apridisk, try (in succession)
	// DSK and EDSK
	if (fmt == 9 && err != 0)
	{
		err = dsk_open(&drv, filename.c_str(), "dsk", NULL);
		if (err) fmt = 2;
	}
	if (fmt == 2 && err != 0)
	{
		err = dsk_open(&drv, filename.c_str(), "edsk", NULL);
		if (err) fmt = 6;
	}
	if (!err)	// It can be opened. Great.
	{
		*type = dsktypes[fmt];
		dsk_close(&drv);
		return UIE_OK; 
	}	
	// File can't be opened. 
	return askCreateDisc(filename, fmt, type);
}


UiEvent QDAE_UI::askCreateDisc(string &filename, int fmt, const char **type)
{
	UiMenu uim(this);
	UiEvent uie;	
	dsk_err_t err;
	DSK_PDRIVER drv;

	string sl1;
	sl1 = "  Cannot open: ";
	sl1 += displayName(filename, 40);
	sl1 += "  ";

	uim.add(new UiLabel(sl1, this));
	uim.add(new UiLabel("", this));
	uim.add(new UiCommand(SDLK_c, "  Create it  ", this));
	uim.add(new UiCommand(SDLK_a, "  Abandon  ", this));
	centreContainer(uim);
	uie = uim.eventLoop();
	if (uie == UIE_QUIT || uie == UIE_CANCEL) return uie;
	if (uim.getKey(uim.getSelected()) != SDLK_c) return UIE_CANCEL;
	// We are at liberty to create it.

	if (!dsktypes[fmt])	// No type was specified
	{
		if (filename == A_DRIVE || filename == B_DRIVE) fmt = 1;
		else						fmt = 2;

		uie = getDiscType(fmt, true);
		if (uie == UIE_QUIT || uie == UIE_CANCEL) return uie;
	}
	// Try to create a new blank disc image
	err = dsk_creat(&drv, filename.c_str(), dsktypes[fmt], NULL);
	if (!err)	// It can be created. Great.
	{
		*type = dsktypes[fmt];
		dsk_close(&drv);
		return UIE_OK; 
	}	
	UiMenu uim2(this);
		
	sl1 = "  Cannot open: ";
	sl1 += displayName(filename, 40);
	sl1 += "  ";

	uim2.add(new UiLabel(sl1, this));
	uim2.add(new UiCommand(SDLK_a, "  Abandon  ", this));
	centreContainer(uim2);
	uie = uim2.eventLoop();
	if (uie == UIE_QUIT || uie == UIE_CANCEL) return uie;

	return UIE_CANCEL;
}

UiEvent QDAE_UI::getDiscType(int &fmt, bool create)
{
	UiEvent uie;
	int inMenu = 1;
	int sel = -1;
	int ofmt = fmt;

	while(inMenu)
	{
		UiMenu uim(this);

		uim.add(new UiLabel  ("      Drive type      ", this));
		uim.add(new UiSeparator(this));
		if (!create)
		{
			uim.add(new UiSetting(SDLK_0, (fmt == 0), "  Auto detect filetype  ", this));
		}
		uim.add(new UiSetting(SDLK_1, (fmt == 1), "  Floppy drive  ", this));
		uim.add(new UiSetting(SDLK_9, (fmt == 9), "  Apridisk image  ", this));
		uim.add(new UiSetting(SDLK_2, (fmt == 2), "  CPCEMU/JOYCE .DSK  ", this));
		uim.add(new UiSetting(SDLK_3, (fmt == 3), "  Raw disc image file  ", this));
//		uim.add(new UiSetting(SDLK_4, (fmt == 4), "  MYZ80 hard drive  ", this));
//		uim.add(new UiSetting(SDLK_5, (fmt == 5), "  .CFI - compressed raw   ", this));
		uim.add(new UiSetting(SDLK_6, (fmt == 6), "  Extended .DSK  ", this));
#ifdef HAVE_WINDOWS_H
		uim.add(new UiSetting(SDLK_7, (fmt == 7), "  Alternative floppy  ", this));
#endif
//		uim.add(new UiSetting(SDLK_8, (fmt == 8), "  Folder  ", this));
		uim.add(new UiSeparator(this));
		uim.add(new UiCommand(SDLK_o, "  OK  ", this));
		uim.add(new UiCommand(SDLK_ESCAPE, "  Cancel  ", this));
		uim.setSelected(sel);
                centreContainer(uim);
		uie = uim.eventLoop();
		if (uie == UIE_QUIT || uie == UIE_CANCEL) 
		{
			fmt = ofmt;
			return uie;
		}
		sel = uim.getSelected();
		switch(uim.getKey(sel))
		{
			case SDLK_0: fmt = 0; break;
			case SDLK_1: fmt = 1; break;
			case SDLK_2: fmt = 2; break;
			case SDLK_3: fmt = 3; break;
			case SDLK_4: fmt = 4; break;
			case SDLK_5: fmt = 5; break;
			case SDLK_6: fmt = 6; break;
			case SDLK_7: fmt = 7; break;
			case SDLK_8: fmt = 8; break;
			case SDLK_9: fmt = 9; break;
			case SDLK_ESCAPE: 
					fmt = ofmt;
					return UIE_CANCEL;
			case SDLK_o: inMenu = 0; break;
			default: break;
		}
	}
	return UIE_OK;
}


/* "Mouse" menu */
UiEvent QDAE_UI::menuF3(int x, int y)
{
	UiMenu uim(this);
	UiEvent uie;

        x = alignX(x);
        y = alignY(y);
	bool capt = m_host->m_video->isMouseCaptured();

	uim.add(new UiSetting(SDLK_c, capt,  "  Capture mouse", this));
	uim.add(new UiSetting(SDLK_r, !capt, "  Release mouse", this));
	uim.setX(x);
	uim.setY(y);

	uie = uim.eventLoop();

	if (uie == UIE_QUIT) return uie;
	switch(uim.getKey(uim.getSelected()))
	{
		case SDLK_c: m_host->m_video->captureMouse(1); break;
		case SDLK_r: m_host->m_video->captureMouse(0); break;
		default: break;
	}
	return UIE_CONTINUE;
}

/* JOYCE, which is the source of all the user-interface code in QDAE, has 
 * an elaborate OO structure, with classes for every device in the system, 
 * each of which may choose to provide a settings screen. The f6 menu then 
 * just iterates over these objects to provide a menu of things with 
 * settings. 
 *
 * The equivalent abilities in QDAE are embryonic at the moment, so I'll
 * content myself with describing .qdaerc.
 */
UiEvent QDAE_UI::menuF6(int x, int y)
{
	return m_host->settingsMenu(this, x, y);
/*
	UiMenu sorry(this);

	string homeFile = displayName(getHomeDir() + "/" RCFILE, 70);

	sorry.add(new UiLabel("Sorry, there is no settings screen for QDAE yet.", this))
 	     .add(new UiLabel("", this))
	     .add(new UiLabel("You must edit the settings manually in the file:", this))
	     .add(new UiLabel(homeFile, this))
	     .add(new UiLabel("", this))
	     .add(new UiLabel("Settings are HOMEDIR=xxxx (set directory for QDAE data files)", this))
	     .add(new UiLabel("and KEYBOARD=xxx (select keyboard layout)", this))

	   .add(new UiLabel("", this))
           .add(new UiCommand(SDLK_o, "  OK  ", this));

	centreContainer(sorry);
	return sorry.eventLoop();
*/
}


UiEvent QDAE_UI::menuF7(int x, int y)
{
	UiMenu uim(this);
	UiEvent uie;

        x = alignX(x);
        y = alignY(y);

	uim.add(new UiCommand (SDLK_r, "  Registers   ", this, UIG_SUBMENU));
	uim.add(new UiCommand (SDLK_d, "  Dump memory  ", this));
	uim.add(new UiCommand (SDLK_t, "  Trace On  ", this));
	uim.add(new UiCommand (SDLK_o, "  Trace Off  ", this));
	uim.setX(x);
	uim.setY(y);

	uie = uim.eventLoop();

	if (uie == UIE_QUIT) return uie;
	switch(uim.getKey(uim.getSelected()))
	{
		case SDLK_r: return registers();
		case SDLK_d: 
		{
			FILE *fp;
			UiEvent uie;

			fp = fopen("apricot.ram", "wb");
			fwrite(m_host->m_mem_base, 1024, 1024, fp);
			fclose(fp);

			uie = report("Memory dumped as apricot.ram", "  OK");
			if (uie == UIE_QUIT) return uie;
		}
		break;
		case SDLK_t: m_host->settings.debug = 3; break;
		case SDLK_o: m_host->settings.debug = 0; break;
		default:
			break;
	}
	return UIE_CONTINUE;
}

void QDAE_UI::showRegs(char buf[4][80])
{
	int n;

	sprintf(buf[0], "AX=%04x BX=%04x CX=%04x DX=%04x SI=%04x DI=%04x SP=%04x BP=%04x",
			m_host->R_AX, m_host->R_BX, m_host->R_CX, 
			m_host->R_DX, m_host->R_SI, m_host->R_DI,
			m_host->R_SP, m_host->R_BP);
	sprintf(buf[1], "IP=%04x CS=%04x DS=%04x ES=%04x SS=%04x ",
			m_host->R_IP, m_host->R_CS, m_host->R_DS,
			m_host->R_ES, m_host->R_SS);

	if (ACCESS_FLAG(m_host,F_OF)) strcat(buf[1],"OV ");     /* CHECKED... */
	else                 strcat(buf[1],"NV ");
	if (ACCESS_FLAG(m_host,F_DF)) strcat(buf[1],"DN ");
	else                 strcat(buf[1],"UP ");
	if (ACCESS_FLAG(m_host,F_IF)) strcat(buf[1],"EI ");
	else                 strcat(buf[1],"DI ");
	if (ACCESS_FLAG(m_host,F_SF)) strcat(buf[1],"NG ");
	else                 strcat(buf[1],"PL ");
	if (ACCESS_FLAG(m_host,F_ZF)) strcat(buf[1],"ZR ");
	else                 strcat(buf[1],"NZ ");
	if (ACCESS_FLAG(m_host,F_AF)) strcat(buf[1],"AC ");
	else                 strcat(buf[1],"NA ");
	if (ACCESS_FLAG(m_host,F_PF)) strcat(buf[1],"PE ");
	else                 strcat(buf[1],"PO ");
	if (ACCESS_FLAG(m_host,F_CF)) strcat(buf[1],"CY ");
	else                 strcat(buf[1],"NC ");

	strcpy(buf[2], "At CS:IP: ");
	for (n = 0; n < 16; n++)
	{
		sprintf(buf[2] + strlen(buf[2]), "%02x ", 
				m_host->fetch(m_host->R_CS, m_host->R_IP + n));
	}	
	strcpy(buf[3], "At SS:SP: ");
	for (n = 0; n < 16; n += 2)
	{
		sprintf(buf[3] + strlen(buf[3]), "%04x  ", 
				m_host->fetch16(m_host->R_SS, m_host->R_SP + n));
	}
}


UiEvent QDAE_UI::registers(void)
{
	UiMenu uim(this);
	char buf[4][80];
	
	showRegs(buf);


        uim.add(new UiLabel(buf[0], this))
	   .add(new UiLabel(buf[1], this))
	   .add(new UiLabel(buf[2], this))
	   .add(new UiLabel(buf[3], this))
	   .add(new UiLabel("", this))
           .add(new UiCommand(SDLK_o, "  OK  ", this));

	centreContainer(uim);
	return uim.eventLoop();
}



UiEvent QDAE_UI::illegal_op(void)
{
	UiMenu uim(this);
	UiEvent uie;

	char buf[4][80];
	
	showRegs(buf);


        uim.add(new UiLabel("                      **   ILLEGAL OPCODE   **  ", this))
	   .add(new UiLabel(buf[0], this))
	   .add(new UiLabel(buf[1], this))
	   .add(new UiLabel(buf[2], this))
	   .add(new UiLabel(buf[3], this))
	   .add(new UiLabel("", this))
           .add(new UiCommand(SDLK_t, "  Terminate QDAE  ", this))
           .add(new UiCommand(SDLK_r, "  Reboot emulated system  ", this))
           .add(new UiCommand(SDLK_a, "  Attempt recovery (DOS)  ", this))
           .add(new UiCommand(SDLK_c, "  Attempt recovery (CP/M) ", this))
	   .add(new UiCommand(SDLK_d, "  Dump memory as file  ", this));

	centreContainer(uim);
	acquireScreen();
	uie = uim.eventLoop();
	releaseScreen();
	if (uie == UIE_QUIT) return uie;
	switch(uim.getKey(uim.getSelected()))
	{
		case SDLK_t: return UIE_QUIT;
		case SDLK_r: m_host->reboot(); return UIE_CONTINUE;
		case SDLK_c:
/* CP/M recovery attempt: Transfer control to INT E0 with CX = DX = 0 */
			/* Set CS:IP to the contents of the INT E0 vector */
			m_host->R_IP = m_host->fetch16(0x380);
			m_host->R_CS = m_host->fetch16(0x382);
			m_host->R_DX = 0;
			m_host->R_CX = 0;
			return UIE_CONTINUE;
/* DOS recovery attempt: INT 21, AX=4C00h */
		case SDLK_a:
			m_host->R_IP = m_host->fetch16(0x84);
			m_host->R_CS = m_host->fetch16(0x86);
			m_host->R_AX = 0x4C00;
			return UIE_CONTINUE;
		case SDLK_d: 
		{
			FILE *fp;
			UiEvent uie;

			fp = fopen("apricot.ram", "wb");
			fwrite(m_host->m_mem_base, 1024, 1024, fp);
			fclose(fp);

			uie = report("Memory dumped as apricot.ram", "  OK");
			if (uie == UIE_QUIT) return uie;
		}
		break;

	}
	return UIE_CONTINUE;
}


UiEvent QDAE_UI::halted(void)
{
	UiMenu uim(this);
	UiEvent uie;

	char buf[4][80];
	
	showRegs(buf);


        uim.add(new UiLabel("                      **   SYSTEM HALTED   **  ", this))
	   .add(new UiLabel(buf[0], this))
	   .add(new UiLabel(buf[1], this))
	   .add(new UiLabel(buf[2], this))
	   .add(new UiLabel(buf[3], this))
	   .add(new UiLabel("", this))
           .add(new UiCommand(SDLK_t, "  Terminate QDAE  ", this))
           .add(new UiCommand(SDLK_r, "  Reboot emulated system  ", this))
           .add(new UiCommand(SDLK_a, "  Attempt recovery (DOS)  ", this))
           .add(new UiCommand(SDLK_c, "  Attempt recovery (CP/M) ", this))
	   .add(new UiCommand(SDLK_d, "  Dump memory as file  ", this));

	centreContainer(uim);
	acquireScreen();
	uie = uim.eventLoop();
	releaseScreen();
	if (uie == UIE_QUIT) return uie;
	switch(uim.getKey(uim.getSelected()))
	{
		case SDLK_t: return UIE_QUIT;
		case SDLK_r: m_host->reboot(); return UIE_CONTINUE;
		case SDLK_c:
/* CP/M recovery attempt: Transfer control to INT E0 with CX = DX = 0 */
			/* Set CS:IP to the contents of the INT E0 vector */
			m_host->R_IP = m_host->fetch16(0x380);
			m_host->R_CS = m_host->fetch16(0x382);
			m_host->R_DX = 0;
			m_host->R_CX = 0;
			return UIE_CONTINUE;
/* DOS recovery attempt: INT 21, AX=4C00h */
		case SDLK_a:
			m_host->R_IP = m_host->fetch16(0x84);
			m_host->R_CS = m_host->fetch16(0x86);
			m_host->R_AX = 0x4C00;
			return UIE_CONTINUE;
		case SDLK_d: 
		{
			FILE *fp;
			UiEvent uie;

			fp = fopen("apricot.ram", "wb");
			fwrite(m_host->m_mem_base, 1024, 1024, fp);
			fclose(fp);

			uie = report("Memory dumped as apricot.ram", "  OK");
			if (uie == UIE_QUIT) return uie;
		}
		break;

	}
	return UIE_CONTINUE;
}



UiEvent QDAE_UI::i8089_error(const char *message)
{
	UiMenu uim(this);
	UiEvent uie;

        uim.add(new UiLabel("          **  8089 error  **  ", this))
	   .add(new UiLabel(message, this))
	   .add(new UiLabel("", this))
           .add(new UiCommand(SDLK_t, "  Terminate QDAE  ", this))
           .add(new UiCommand(SDLK_r, "  Reboot emulated system  ", this))
           .add(new UiCommand(SDLK_i, "  Ignore error  ", this));

	centreContainer(uim);
	acquireScreen();
	uie = uim.eventLoop();
	releaseScreen();
	if (uie == UIE_QUIT) return uie;
	switch(uim.getKey(uim.getSelected()))
	{
		case SDLK_t: return UIE_QUIT;
		case SDLK_r: m_host->reboot(); return UIE_CONTINUE;
	}
	return UIE_CONTINUE;
}


