/*************************************************************************
 
    Quick And Dirty Apricot Emulator
    Copyright (C) 2005  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/


#include "qdae.hxx"
#include "sdl_video.hxx"
#include "qdae_ui.hxx"
#include "wd2797.hxx"


#define FDC_DEBUGLEVEL 1
//#define FDC_DEBUGLEVEL 5

#define FDC_TRACE(l, s) do { if (l < FDC_DEBUGLEVEL) \
			{ printf("FDC: "); printf s; putchar('\n'); \
		 	fflush(stdout); } } while(0)

WD2797::WD2797(PC_ENV *m) : Device("wd2797")
{
	int n;

	pc_env = m;
	busy = 0;
	drive = 0;
	motor = 0;
	irq_flags = 0;	

	for (n = 0; n < MAX_FLOPPIES; n++)
	{
		dsk_drive[n] = NULL;
	}
	reset();
}

WD2797::~WD2797()
{
	int n;

	for (n = 0; n < MAX_FLOPPIES; n++)
	{
		eject(n);
	}
}

void WD2797::eject(int drive)
{
	if (dsk_drive[drive])
	{
		dsk_close(&dsk_drive[drive]);
	}
	*floppy_filename[drive] = "";
	floppy_type[drive] = NULL;
}


UiEvent WD2797::insert(int drive, string filename, const char *type)
{
	dsk_err_t err;
	DSK_PDRIVER pdrv;
	DSK_GEOMETRY geom;

	err = dsk_open(&pdrv, filename.c_str(), type, NULL);
	if (!err) err = dsk_getgeom(pdrv, &geom);
	if (err)
	{
		FDC_TRACE(0,("Failed to open %s: %s", filename.c_str(),
						dsk_strerror(err)));
		return gl_ui->reportV("Failed to open %s: %s", "  OK  ",
				filename.c_str(), dsk_strerror(err));
	}
	*floppy_filename	[drive] = filename;
	floppy_type     [drive] = type;
	dsk_drive       [drive] = pdrv;
	dsk_geom        [drive] = geom;
	return UIE_CONTINUE;	
}



void WD2797::xlt_dsk_error(dsk_err_t err)
{
	fdc_reg[0] &= 0xA3; 
	switch(err)
	{
		case DSK_ERR_OK:      fdc_reg[0] &= 3; break;
		case DSK_ERR_NOTRDY:  fdc_reg[0] |= 0x80; break;
		case DSK_ERR_RDONLY:  fdc_reg[0] |= 0x40; break;
		case DSK_ERR_NOADDR:  fdc_reg[0] |= 0x10; break;
		case DSK_ERR_DATAERR: fdc_reg[0] |= 0x08; break;
		default:
		case DSK_ERR_NODATA:  fdc_reg[0] |= 0x04; break;
	}
}

void WD2797::read_sector(unsigned char cmd_byte)
{
	int xcyl = fdc_reg[1];
	int xsec = fdc_reg[2];
	int xhead = (cmd_byte & 2) ? 1 : 0;
	int deleted = 0;
	dsk_err_t err;

	FDC_TRACE(4,("Reading sector. Expected cyl=%d head=%d sec=%d. "
			"Reading at cylinder %d", xcyl, xhead, xsec,
			cur_cyl[drive]));
	if (dsk_geom[drive].dg_secsize > 8192)
	{	
		// Avoid buffer overflow on read
		xlt_dsk_error(DSK_ERR_NOADDR);
		return;
	}
	dsk_geom[drive].dg_noskip = 1;
	err = dsk_xread(dsk_drive[drive], &dsk_geom[drive],
			transfer_buf, 
			cur_cyl[drive], xhead, xcyl, xhead, xsec,
			dsk_geom[drive].dg_secsize, &deleted);

	if (err == DSK_ERR_NOTIMPL)
	{
		FDC_TRACE(4,("No dsk_xread, falling back on dsk_pread"));
		if (cur_cyl[drive] != xcyl)
		{
			err = DSK_ERR_NOADDR;
		}
		else
		{
			err = dsk_pread(dsk_drive[drive], &dsk_geom[drive],
				transfer_buf, 
				cur_cyl[drive], xhead, xsec);
		}
	}
	FDC_TRACE(4,("Read sector, result=%d", err));
	xlt_dsk_error(err);
	if (deleted) fdc_reg[0] |= 0x20;
	else	     fdc_reg[0] &= ~0x20;

	if (err == DSK_ERR_OK)
	{
		transfer_count = dsk_geom[drive].dg_secsize;
		transfer_ptr = 0;	
	}
	busy = 5;
}

void WD2797::write_sector(unsigned char cmd_byte)
{
	FDC_TRACE(4,("Writing sector: Begin"));

	write_cmd = cmd_byte;
	busy = 5;
	transfer_count = dsk_geom[drive].dg_secsize;
	transfer_ptr = 0;	
	if (floppy_readonly[drive])
	{
		xlt_dsk_error(DSK_ERR_RDONLY);
		return;
	}
}

void WD2797::write_sector_end()
{
	int xcyl = fdc_reg[1];
	int xsec = fdc_reg[2];
	int xhead = (write_cmd & 2) ? 1 : 0;
	int deleted = 0;
	dsk_err_t err;

	FDC_TRACE(4,("Writing sector cyl=%d head=%d sec=%d", xcyl, xhead, xsec));
	if (floppy_readonly[drive])
	{
		xlt_dsk_error(DSK_ERR_RDONLY);
		return;
	}
	dsk_geom[drive].dg_noskip = 1;
	err = dsk_xwrite(dsk_drive[drive], &dsk_geom[drive],
			transfer_buf, 
			cur_cyl[drive], xhead, xcyl, xhead, xsec,
			dsk_geom[drive].dg_secsize, deleted);

	if (err == DSK_ERR_NOTIMPL)
	{
		err = dsk_pwrite(dsk_drive[drive], &dsk_geom[drive],
			transfer_buf, 
			cur_cyl[drive], xhead, xsec);
	}
	FDC_TRACE(4,("Write sector, result=%d", err));
	xlt_dsk_error(err);
}


void WD2797::probe_devices()
{
	int n;
	dsk_err_t err;

	for (n = 0; n < MAX_FLOPPIES; n++)
	{
		if (floppy_filename[n][0] != "" && !dsk_drive[n])
		{
			err = dsk_open(&dsk_drive[n], 
					floppy_filename[n]->c_str(), 
					floppy_type[n], NULL);
			if (!err) err = dsk_getgeom(dsk_drive[n], &dsk_geom[n]);
			if (err)
			{
				UiEvent e = gl_ui->reportV(
						"Failed to open %s: %s",
						"  OK  ",
						floppy_filename[n]->c_str(), 
						dsk_strerror(err));
				FDC_TRACE(0,("Failed to open %s: %s",
						floppy_filename[n]->c_str(), 
						dsk_strerror(err)));
				floppy_filename[n][0] = "";
				if (e == UIE_QUIT) 
				{
					exit(0);	
				}
			}
		}	
	}
}

void WD2797::begin_command(u_int8 value)
{
	switch(value & 0xF0)
	{
		case 0x00: FDC_TRACE(4,("Restore")); 
			   cur_cyl[drive] = 0;		    
			   if (value & 0x10) fdc_reg[1] = cur_cyl[drive];
			   if (cur_cyl[drive] == 0) fdc_reg[0] |= 4;
			   else				fdc_reg[0] &= ~4;
			   busy = 5;
			   break;
		case 0x10: FDC_TRACE(4,("Seek %d", fdc_reg[3])); 
			   cur_cyl[drive] = fdc_reg[3];		    
			   if (value & 0x10) fdc_reg[1] = cur_cyl[drive];
			   if (cur_cyl[drive] == 0) fdc_reg[0] |= 4;
			   else				fdc_reg[0] &= ~4;
			   busy = 5;
			   break;
		case 0x20: 
		case 0x30: FDC_TRACE(4,("Step UNIMPLEMENTED")); break;
		case 0x40: 
		case 0x50: FDC_TRACE(4,("Step in")); 
			   cur_cyl[drive]++;		    
			   if (value & 0x10) fdc_reg[1] = cur_cyl[drive];
			   if (cur_cyl[drive] == 0) fdc_reg[0] |= 4;
			   else				fdc_reg[0] &= ~4;
			   busy = 5;
			   break;
		case 0x60: 
		case 0x70: FDC_TRACE(4,("Step out")); 
			   if (cur_cyl[drive] > 0) cur_cyl[drive]--;
			   if (value & 0x10) fdc_reg[1] = cur_cyl[drive];
			   if (cur_cyl[drive] == 0) fdc_reg[0] |= 4;
			   else				fdc_reg[0] &= ~4;
			   busy = 5;
			   break;
		case 0x80: 
		case 0x90: FDC_TRACE(4,("Read Sector 0x%02x 0x%02x 0x%02x",
					value & 0x1F,
					fdc_reg[1],
					fdc_reg[2])); 
			   read_sector(value & 0x1F);	   
			   break;
		case 0xA0: 
		case 0xB0: FDC_TRACE(4,("Write Sector")); 
			   write_sector(value & 0x1F);	   
			   break;
		case 0xC0: FDC_TRACE(4,("Read Address UNIMPLEMENTED")); break;
		case 0xD0: FDC_TRACE(4,("Force Interrupt 0x%02x", value & 0x0F)); 
			   irq_flags = value & 0x0F;
			   stop();
			   break;
		case 0xE0: FDC_TRACE(4,("Read Track UNIMPLEMENTED")); break;
		case 0xF0: FDC_TRACE(4,("Write Track UNIMPLEMENTED")); break;
	}
}

void WD2797::write_command(u_int8 value)
{
	begin_command(value);
}


void WD2797::select_drive(int d)
{
	drive = d;
	FDC_TRACE(5,("Select drive %d", d));
}

void WD2797::load_head(bool b)
{
	FDC_TRACE(5,("Drive %d Load head 0x%02x", drive, b));
	floppy_indicator[drive] = b;
	gl_ledsdirty = 1;
}

void WD2797::set_motor(u_int8 value)
{
	motor = value;
	FDC_TRACE(5,("Set motor 0x%02x", value));
	if (motor) probe_devices();
}

void WD2797::write(u_int16 reg, u_int8 value)
{
	FDC_TRACE(5,("Write register %d value=0x%02x", reg, value));
	switch(reg)
	{
		case 0: // Command
			write_command(value);
			break;
		case 3:	if (transfer_count)
			{
				transfer_buf[transfer_ptr++] = value;
				--transfer_count;
				if (!transfer_count)
				{
					write_sector_end();
				}
			}
			// FALL THROUGH
		default: // Track / sector / data reg
			fdc_reg[reg] = value;
			break;
	}
}	

u_int8 WD2797::read(u_int16 reg)
{
	u_int8 value = fdc_reg[reg];

	if (reg == 0)	/* Status */
	{
		value = fdc_reg[0];
		if (busy)
		{
			--busy;
			value |= 1;	/* Busy */
		}
		if (transfer_count) 
		{
			// value |= 1;	// ? Needed ?
		}
		if (dsk_drive[drive] == NULL)
		{
			value |= 0x80;	/* Not ready */
		}
	}
	if (reg == 3 && transfer_count)
	{
		value = transfer_buf[transfer_ptr++];
		--transfer_count;
	}
	FDC_TRACE(5,("Read register %d returns 0x%02x", reg, value));

	return value;
}

void WD2797::stop()
{
	busy = 0;
	transfer_count = 0;
}


void WD2797::reset()
{
	int n;

	for (n = 0; n < 4; n++)
	{
		fdc_reg[n] = 0;
	}
	drive = 0;
	motor = 0;
	for (n = 0; n < MAX_FLOPPIES; n++)
	{
		cur_cyl[n] = 0;
		if (dsk_drive[n]) dsk_close(&dsk_drive[n]);
	}
	stop();
}


bool WD2797::parseNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	return true;
}

bool WD2797::storeNode(xmlDoc *doc, xmlNs *ns, xmlNode *node)
{
	return true;
}


