/*************************************************************************
 
    Quick And Dirty Apricot Emulator
    Copyright (C) 2005  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "qdae.hxx"
#include "device.hxx"
#include "z80_sio.hxx"
#include "i8259.hxx"

//#define SIO_TRACE(ch,s) do { printf("Z80 SIO(%d): ",ch+1); printf s; \
//	putchar('\n'); fflush(stdout); } while(0)

#define SIO_TRACE(ch, s)



Z80_SIO::Z80_SIO(PC_ENV *m) : Device("z80sio")
{
	int n;

	irq_base = 0;
	last_z80_op = 0x00;
	m_host = m;
	m_pic = NULL;
	m_picirq = -1;
	sio_ch[0].m_listener = sio_ch[1].m_listener = NULL;
	reset();
}

Z80_SIO::~Z80_SIO()
{

}

static u_int8 st_ir[16] = 
{
	0x80, 0x07, 0x19, 0x9e, 
	0x2a, 0xad, 0xb3, 0x34,
	0xcb, 0x4c, 0x52, 0xd5,
	0x61, 0xe6, 0xf8, 0x7f
};

/* Break infrared word into 3 apricot nibbles. It seems to use the strange
 * middle-endian ordering below. */
void Z80_SIO::write_ir(int ch, u_int16 irword)
{
	u_int8 nibble[3];
	int n;

	nibble[0] = (irword >> 8) & 0x0F;
	nibble[1] = irword & 0x0F;
	nibble[2] = (irword >> 4) & 0x0F;

	for (n = 0; n < 3; n++)
	{
		write_rx(ch, st_ir[nibble[n]]);
	}
}

void Z80_SIO::write_rx(int ch, u_int8 value)
{
	sio_ch[ch].write_rx(value);
}

void SerialChannel::write_rx(u_int8 value)
{
/* Buffer full... */
	if (rx_qhead == rx_qtail - 1 ||
	    rx_qhead == 511 && rx_qtail == 0) return;

	rx_queue[rx_qhead++] = value;
	if (rx_qhead >= sizeof(rx_queue)) rx_qhead = 0;
}


u_int8 SerialChannel::read_rx()
{
	u_int8 value;

	if (rx_qhead == rx_qtail) return 0xFF;

	value = rx_queue[rx_qtail++];
	if (rx_qtail >= sizeof(rx_queue)) rx_qtail = 0;
	return value;
}

u_int8 SerialChannel::read_tx()
{
	u_int8 value;

	if (tx_qhead == tx_qtail) return 0xFF;

	value = tx_queue[tx_qtail++];
	if (tx_qtail >= sizeof(tx_queue)) tx_qtail = 0;
	return value;
}



void SerialChannel::write_tx(u_int8 value)
{
/* Buffer full... */
	if (tx_qhead == tx_qtail - 1 ||
	    tx_qhead == 511 && tx_qtail == 0) return;

	tx_queue[tx_qhead++] = value;
	if (tx_qhead >= sizeof(tx_queue)) tx_qhead = 0;

	if (m_listener)
	{
		m_listener->receive(this);
	}
}




void SerialChannel::command(int cmd)
{
	switch(cmd & 7)
	{
		case 0: // Nop
			break;
		case 3: // Channel reset
			SIO_TRACE(number, ("Reset"));
			break;
		case 4: // Enable interrupt on RX char
			SIO_TRACE(number, ("Enable RX interrupt"));
			rx_interrupt = 1;
			break;
		default:
			SIO_TRACE(number,("SIO command %d UNIMPLEMENTED", cmd));
       			break;
	}
}

void Z80_SIO::opcode(u_int8 value)
{
	if (last_z80_op == 0xED && value == 0x4B)
	{
		iret();
	}
	last_z80_op = value;
}


void Z80_SIO::iret()
{
	SIO_TRACE(0, ("Return from interrupt"));
	if (m_pic)
	{
		m_pic->setirq(m_picirq, false);
	}
}

void Z80_SIO::write(u_int16 port, u_int8 value)
{
	int channel;
	SerialChannel *ch;

	if (port < 2) channel = 0;
	else	      channel = 1;

	ch = &sio_ch[channel];

	if (port & 1) 
	{
		ch->reg[ch->reg_sel] = value;
		if (ch->reg_sel == 0)
		{
			ch->reg_sel = value & 7;
			if ((value >> 3) == 7)
			{
				iret();
			}
			else
			{
				ch->command(value >> 3);
			}
		}
		else
		{
			ch->reg_sel = 0;
		}
	}
	else
	{
		ch->write_tx(value);
	}
	SIO_TRACE(channel,("Write port %04x value=0x%02x", port, value));
}	


u_int8 Z80_SIO::read(u_int16 port, bool locked)
{
	int channel;
	SerialChannel *ch;

	u_int8 value = 0xFF;
	if (port < 2) channel = 0;
	else	      channel = 1;
	ch = &sio_ch[channel];
      
	if (port & 1)
	{
		SIO_TRACE(channel,("Read port %02x; reg=%d", port,
					ch->reg_sel));
		switch(ch->reg_sel)
		{
			case 2: value = ch->reg[2]; // Interrupt vector
				break;
			case 0: // Status 0
				value = 0;
				if (ch->rx_qhead != ch->rx_qtail) value |= 1;
				if (ch->tx_qhead == ch->tx_qtail) value |= 4;
				// XXX 6 more bits to go...
				break;
			case 1: // Status 1
				// XXX Wrong
			default:
				value = 0xFF;
				break;
		}
	}
	else
	{
/* For the Xi: The bus lock signal is connected to the Z80-SIO's M1 line, so 
 * a read with the bus locked is treated as part of an interrupt acknowledge
 * cycle; the interrupt is reset and the last IRQ returned. */
		if (locked)
		{
			if (m_pic) m_pic->setirq(m_picirq, false);
			value = m_lastirq;
			SIO_TRACE(channel, ("Read port %04x returns last IRQ %d\n",
					port, value));
			m_lastirq = -1;
		}
		else
		{
			value = ch->read_rx();
		}
	}
	SIO_TRACE(channel,("Read port %04x returns 0x%02x", port, value));

	return value;
}

void Z80_SIO::reset()
{
	int n;
	for (n = 0; n < 2; n++)
	{
		sio_ch[n].number = n;
		sio_ch[n].reset();
	}
	m_raising = false;
	m_lastirq = -1;
}

void SerialChannel::reset()
{
	reg_sel = 0;
	tx_qhead = 0;
	tx_qtail = 0;
	rx_qhead = 0;
	rx_qtail = 0;
	isr_countdown = 0;
}


u_int8 Z80_SIO::tick()
{
	int n;
	if (!ACCESS_FLAG(m_host, F_IF)) return 0;

	for (n = 0; n < 2; n++)
	{
		if (sio_ch[n].isr_countdown)
		{
			sio_ch[n].isr_countdown--;
			continue;
		}
/* If we've been asked to interrupt, we're interrupting, and there's no 
 * character available, then stop interrupting */
		if (sio_ch[n].rx_qhead == sio_ch[n].rx_qtail 
		&&  (sio_ch[n].reg[1] & 0x18) && m_raising)
		{
			m_raising = false;
			if (m_pic) m_pic->setirq(m_picirq, false);
		}

		if (sio_ch[n].rx_qhead != sio_ch[n].rx_qtail 
		&&  (sio_ch[n].reg[1] & 0x18))
		{
			m_lastirq = sio_ch[1].reg[2] + ((1-n)*8) + 4;
			m_raising = true;
			if (m_pic)
			{
				SIO_TRACE(n, ("Raising interrupt 0x%02x via PIC: m_lastirq:=%d", m_picirq, m_lastirq));
				m_pic->setirq(m_picirq, true);
//				m_host->settings.debug |= DEBUG_DECODE_F | DEBUG_TRACE_F;
			}
			else
			{
	
				SIO_TRACE(n, ("Raising interrupt 0x%02x", m_lastirq));
				m_host->i86_intr_raise(this);
			}
// And don't interrupt again for 2000 instructions, to give the CPU time
// to deal with this one.
			sio_ch[n].isr_countdown = 2000;
		}
	}
	return 0;
}


bool Z80_SIO::parseNode(xmlDoc *doc, xmlNs *ns, xmlNode *cur)
{
	return true;
}

bool Z80_SIO::storeNode(xmlDoc *doc, xmlNs *ns, xmlNode *cur)
{
	return true;
}


void Z80_SIO::setListener(int ch, SerialListener *l)
{
	if (ch >= 0 && ch < 2)
	{
		sio_ch[ch].m_listener = l;
	}
}

u_int8 Z80_SIO::ackInterrupt()
{
	u_int8 result = m_lastirq;
// Seems not to work if you do this on the Rascal
	m_lastirq = -1;
	return result;
}
